/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * @(#)AccessBridgeATInstance.cpp	1.11 02/08/23
 */

/*
 * A class to track key AT instance info from the JavaAccessBridge
 */

#ifdef _DEBUG
	#define DEBUGGING_ON
#endif

#include "AccessBridgeDebug.h"
#include "AccessBridgeATInstance.h"
#include "AccessBridgeMessages.h"

#include <windows.h> 
#include <winbase.h>


/**
 *
 *
 */
AccessBridgeATInstance::AccessBridgeATInstance(HWND ourABWindow, HWND winABWindow,
											   char *memoryFilename,
											   AccessBridgeATInstance *next) {
	ourAccessBridgeWindow = ourABWindow;
	winAccessBridgeWindow = winABWindow;
	nextATInstance = next;
	javaEventMask = 0;
	accessibilityEventMask = 0;
	strncpy(memoryMappedFileName, memoryFilename, cMemoryMappedNameSize);
}

/**
 *
 *
 */
AccessBridgeATInstance::~AccessBridgeATInstance() {
	PrintDebugString("\r\nin AccessBridgeATInstance::~AccessBridgeATInstance"); 

		// if IPC memory mapped file view is valid, unmap it
	if (memoryMappedView != (char *) 0) {
		PrintDebugString("  unmapping memoryMappedView; view = %X", memoryMappedView); 
		UnmapViewOfFile(memoryMappedView);
		memoryMappedView = (char *) 0;
	}
		// if IPC memory mapped file handle map is open, close it
	if (memoryMappedFileMapHandle != (HANDLE) 0) {
		PrintDebugString("  closing memoryMappedFileMapHandle; handle = %X", memoryMappedFileMapHandle); 
		CloseHandle(memoryMappedFileMapHandle);
		memoryMappedFileMapHandle = (HANDLE) 0;
	}
}

/**
 * initiateIPC - sets up the memory-mapped file to do IPC messaging 
 *				 1 files is created: to handle requests for information
 *				 initiated from Windows AT.  The package is placed into
 *				 the memory-mapped file (char *memoryMappedView),
 *				 and then a special SendMessage() is sent.  When the
 *				 JavaDLL returns from SendMessage() processing, the
 *				 data will be in memoryMappedView.  The SendMessage()
 *				 return value tells us if all is right with the world.
 *
 *				 The set-up proces involves creating the memory-mapped
 *				 file, and writing a special string to it so that the
 *				 WindowsDLL so it knows about it as well.
 *
 */
LRESULT
AccessBridgeATInstance::initiateIPC() {
	DWORD errorCode;

	PrintDebugString("\r\nin AccessBridgeATInstance::initiateIPC()"); 
	
	// open Windows-initiated IPC filemap & map it to a ptr

	memoryMappedFileMapHandle = OpenFileMapping(FILE_MAP_READ | FILE_MAP_WRITE,
												FALSE, memoryMappedFileName);
	if (memoryMappedFileMapHandle == NULL) {
		errorCode = GetLastError();
		PrintDebugString("  Failed to CreateFileMapping for %s, error: %X", memoryMappedFileName, errorCode); 
		return errorCode;
	} else {
		PrintDebugString("  CreateFileMapping worked - filename: %s", memoryMappedFileName); 
	}

	memoryMappedView = (char *) MapViewOfFile(memoryMappedFileMapHandle,
											  FILE_MAP_READ | FILE_MAP_WRITE,
											  0, 0, 0);
	if (memoryMappedView == NULL) { 
		errorCode = GetLastError();
		PrintDebugString("  Failed to MapViewOfFile for %s, error: %X", memoryMappedFileName, errorCode); 
		return errorCode;
	} else {
		PrintDebugString("  MapViewOfFile worked - view: %X", memoryMappedView); 
	}


		// look for the JavaDLL's answer to see if it could read the file
	if (strcmp(memoryMappedView, AB_MEMORY_MAPPED_FILE_OK_QUERY) != 0) {
		PrintDebugString("  JavaVM failed to write to memory mapped file %s",
			  memoryMappedFileName); 
		return -1;
	} else {
		PrintDebugString("  JavaVM successfully wrote to file!"); 
	}


		// write some data to the memory mapped file for WindowsDLL to verify
	strcpy(memoryMappedView, AB_MEMORY_MAPPED_FILE_OK_ANSWER);


	return 0;
}

/**
 * sendJavaEventPackage - uses SendMessage(WM_COPYDATA) to do 
 *					      IPC messaging with the Java AccessBridge DLL
 *					      to propogate events to those ATs that want 'em
 *
 */
LRESULT
AccessBridgeATInstance::sendJavaEventPackage(char *buffer, int bufsize, long eventID) {

	PrintDebugString("AccessBridgeATInstance::sendJavaEventPackage() eventID = %X", eventID);
	PrintDebugString("AccessBridgeATInstance::sendJavaEventPackage() (using PostMessage) eventID = %X", eventID);

	if (eventID & javaEventMask) {
		COPYDATASTRUCT toCopy;
		toCopy.dwData = 0;		// 32-bits we could use for something...
		toCopy.cbData = bufsize;
		toCopy.lpData = buffer;

/*
        // try using PostMessage; problem with call coming in in middle of 
        // a call initiated from AT client (e.g. while we're processing
        // a call to getAccessibleContextInfo()
*/		return SendMessage(winAccessBridgeWindow, WM_COPYDATA,
						  (WPARAM) ourAccessBridgeWindow, (LPARAM) &toCopy);

/*        return PostMessage(winAccessBridgeWindow, WM_COPYDATA,
						  (WPARAM) ourAccessBridgeWindow, (LPARAM) &toCopy);
*/	} else {
		return -1;
	}
}


/**
 * sendAccessibilityEventPackage - uses SendMessage(WM_COPYDATA) to do 
 *					  IPC messaging with the Java AccessBridge DLL
 *					  to propogate events to those ATs that want 'em
 *
 */
LRESULT
AccessBridgeATInstance::sendAccessibilityEventPackage(char *buffer, int bufsize, long eventID) {

	PrintDebugString("AccessBridgeATInstance::sendAccessibilityEventPackage() eventID = %X", eventID);

	if (eventID & accessibilityEventMask) {
		COPYDATASTRUCT toCopy;
		toCopy.dwData = 0;		// 32-bits we could use for something...
		toCopy.cbData = bufsize;
		toCopy.lpData = buffer;

/*
        // try using PostMessage; problem with call coming in in middle of 
        // a call initiated from AT client (e.g. while we're processing
        // a call to getAccessibleContextInfo()
*/		return SendMessage(winAccessBridgeWindow, WM_COPYDATA,
						  (WPARAM) ourAccessBridgeWindow, (LPARAM) &toCopy);
/*
		return PostMessage(winAccessBridgeWindow, WM_COPYDATA,
						  (WPARAM) ourAccessBridgeWindow, (LPARAM) &toCopy);
*/	} else {
		return -1;
	}
}


/**
 * findABATInstanceFromATHWND - walk through linked list from 
 *								where we are.  Return the  
 *								AccessBridgeATInstance 
 *								of the ABATInstance that 
 *								matches the passed in vmID;  
 *								no match: return 0
 */
AccessBridgeATInstance *
AccessBridgeATInstance::findABATInstanceFromATHWND(HWND window) {
	// no need to recurse really
	if (winAccessBridgeWindow == window) {
		return this;
	} else {
		AccessBridgeATInstance *current = nextATInstance;
		while (current != (AccessBridgeATInstance *) 0) {
			if (current->winAccessBridgeWindow == window) {
				return current;
			}
			current = current->nextATInstance;
		}
	}
	return (AccessBridgeATInstance *) 0;
}


