/*
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * %W% %E%
 */

/* 
 * A sample Assistive Technology which queries the JavaVM (if any) underneath
 * the mouse pointer to get the Java Accessibility information available
 * for the Java UI object underneath the mouse, using the Java 
 * Accessibility Bridge, AccessBridge.DLL
 */

#include <windows.h>   // includes basic windows functionality
#include <stdio.h>
#include <jni.h>
#include "ferretResource.h"
#include "AccessBridgeCalls.h"
#include "AccessBridgeCallbacks.h"

#define TIMER_ID 1
#define DISPLAY_INFO_MESSAGE WM_USER+1
#define DISPLAY_HWND_INFO_MESSAGE WM_USER+2

#define TEXTBUFSIZE 8192


HWND theDialogWindow;
HINSTANCE theInstance;
BOOL theAccessBridgeLoadedFlag;

BOOL InitWindow (HANDLE hInstance);
BOOL APIENTRY FerretDialogProc (HWND hDlg, UINT message, UINT wParam, LONG lParam);
LRESULT CALLBACK KeyboardProc (int code, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK MouseProc (int code, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK TimerProc(HWND hWnd, UINT uTimerMsg, UINT uTimerID, DWORD dwTime);

void HandleJavaPropertyChange(long vmID, PropertyChangeEvent event,
							  AccessibleContext ac,
							  wchar_t *name, wchar_t *oldValue, wchar_t *newValue);

void HandleJavaFocusGained(long vmID, FocusEvent event, AccessibleContext ac);
void HandleJavaFocusLost(long vmID, FocusEvent event, AccessibleContext ac);

void HandleJavaCaretUpdate(long vmID, CaretEvent event, AccessibleContext ac);

void HandleMouseClicked(long vmID, MouseEvent event, AccessibleContext ac);
void HandleMouseEntered(long vmID, MouseEvent event, AccessibleContext ac);
void HandleMouseExited(long vmID, MouseEvent event, AccessibleContext ac);
void HandleMousePressed(long vmID, MouseEvent event, AccessibleContext ac);
void HandleMouseReleased(long vmID, MouseEvent event, AccessibleContext ac);

void HandleMenuCanceled(long vmID, MenuEvent event, AccessibleContext ac);
void HandleMenuDeselected(long vmID, MenuEvent event, AccessibleContext ac);
void HandleMenuSelected(long vmID, MenuEvent event, AccessibleContext ac);
void HandlePopupMenuCanceled(long vmID, MenuEvent event, AccessibleContext ac);
void HandlePopupMenuWillBecomeInvisible(long vmID, MenuEvent event, AccessibleContext ac);
void HandlePopupMenuWillBecomeVisible(long vmID, MenuEvent event, AccessibleContext ac);

void HandlePropertyNameChange(long vmID, PropertyChangeEvent event, AccessibleContext ac, 
                              wchar_t *oldName, wchar_t *newName);
void HandlePropertyDescriptionChange(long vmID, PropertyChangeEvent event, AccessibleContext ac, 
                                     wchar_t *oldDescription, wchar_t *newDescription);
void HandlePropertyStateChange(long vmID, PropertyChangeEvent event, AccessibleContext ac, 
                               wchar_t *oldState, wchar_t *newState);
void HandlePropertyValueChange(long vmID, PropertyChangeEvent event, AccessibleContext ac, 
                               wchar_t *oldValue, wchar_t *newValue);
void HandlePropertySelectionChange(long vmID, PropertyChangeEvent event, AccessibleContext ac);
void HandlePropertyTextChange(long vmID, PropertyChangeEvent event, AccessibleContext ac);
void HandlePropertyCaretChange(long vmID, PropertyChangeEvent event, AccessibleContext ac, 
                               int oldPosition, int newPosition);
void HandlePropertyVisibleDataChange(long vmID, PropertyChangeEvent event, AccessibleContext ac);
void HandlePropertyChildChange(long vmID, PropertyChangeEvent event, AccessibleContext ac, 
                               jobject oldChild, jobject newChild);
void HandlePropertyActiveDescendentChange(long vmID, PropertyChangeEvent event, 
                                          AccessibleContext ac, 
                                          jobject oldActiveDescendent,
                                          jobject newActiveDescendent);

void HandlePropertyTableModelChange(long vmID, PropertyChangeEvent event, 
                                          AccessibleContext ac, 
                                          wchar_t *oldValue, wchar_t *newValue);

HHOOK prevKbdHook;
HHOOK prevMouseHook;

BOOL updateMouse;
BOOL updateF1;
BOOL updateF2;

BOOL trackMouse;
BOOL trackFocus;
BOOL trackCaret;

BOOL trackMenuSelected;
BOOL trackMenuDeselected;
BOOL trackMenuCanceled;

BOOL trackPopupVisible;
BOOL trackPopupInvisible;
BOOL trackPopupCanceled;

//BOOL trackPropertyChange;

BOOL trackPropertyNameChange;
BOOL trackPropertyDescriptionChange;
BOOL trackPropertyStateChange;
BOOL trackPropertyValueChange;
BOOL trackPropertySelectionChange;
BOOL trackPropertyTextChange;
BOOL trackPropertyCaretChange;
BOOL trackPropertyVisibleDataChange;
BOOL trackPropertyChildChange;
BOOL trackPropertyActiveDescendentChange;
BOOL trackPropertyTableModelChange;


/**
 * WinMain
 *
 */
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nShowCmd) {
	MSG msg;

	theInstance = hInst;

	updateF1 = FALSE;
	updateF2 = FALSE;
	updateMouse = FALSE;

    trackMouse = FALSE;
	trackFocus = FALSE;
	trackCaret = FALSE;
	trackMenuSelected = FALSE;
	trackMenuDeselected = FALSE;
	trackMenuCanceled = FALSE;
	trackPopupVisible = FALSE;
	trackPopupInvisible = FALSE;
	trackPopupCanceled = FALSE;

	trackPropertyNameChange = FALSE;
	trackPropertyDescriptionChange = FALSE;
	trackPropertyStateChange = FALSE;
	trackPropertyValueChange = FALSE;
	trackPropertySelectionChange = FALSE;
	trackPropertyTextChange = FALSE;
	trackPropertyCaretChange = FALSE;
	trackPropertyVisibleDataChange = FALSE;
	trackPropertyChildChange = FALSE;
	trackPropertyActiveDescendentChange = FALSE;
	trackPropertyTableModelChange = FALSE;


    theAccessBridgeLoadedFlag = FALSE;


	if (InitWindow(hInst)) {
		if (initializeAccessBridge() == TRUE) {
            theAccessBridgeLoadedFlag = TRUE;
			while (GetMessage(&msg, NULL, 0, 0)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
            if (theAccessBridgeLoadedFlag == TRUE) {
			    shutdownAccessBridge();
            }
		}
	}
	return(0);
}


/**
 * Real(tm) MS-Windows window initialization
 *
 */
BOOL InitWindow (HANDLE hInstance) {
	theDialogWindow = CreateDialog (hInstance,	"FERRETWINDOW", NULL, FerretDialogProc);

	// If window could not be created, return "failure".
	if (!theDialogWindow)
	   return (FALSE);

	// Make the window visible, update its client area, & return "success".
	SetWindowText (theDialogWindow, "Java Ferret");
	ShowWindow (theDialogWindow, SW_SHOWNORMAL);
	UpdateWindow (theDialogWindow); 
	
	return (TRUE);      
}


/**
 * Gather the Accessible information about the object under the mouse
 */
char *getAccessibleInfo(long vmID, AccessibleContext ac,
						int x, int y, char *buffer, int bufsize) {
    BOOL result;
    char s[1024];
    AccessBridgeVersionInfo versionInfo;
    AccessibleContext parentAC;
	AccessibleContextInfo info;
	AccessibleContextInfo parentInfo;
	AccessibleTextInfo textInfo;
	AccessibleTextItemsInfo textItems;
	AccessibleTextSelectionInfo textSelection;
	AccessibleTextRectInfo rectInfo;
    AccessibleTextAttributesInfo attributeInfo;
	wchar_t tempBuf[1024];
    int selCount;
    int i, j;
    AccessibleContext selectedAC;
    int start;
    int end;
	AccessibleContextInfo relInfo;

    strcpy(buffer, "Version Information:");
    result = GetVersionInfo(vmID, &versionInfo);
    if (result == TRUE) {
        wsprintf(s, "\r\n  Java virtual machine version: %ls", versionInfo.VMversion);
            strncat(buffer, s, bufsize);
        wsprintf(s, "\r\n  Access Bridge Java class version: %ls", versionInfo.bridgeJavaClassVersion);
            strncat(buffer, s, bufsize);
        wsprintf(s, "\r\n  Access Bridge Java DLL version: %ls", versionInfo.bridgeJavaDLLVersion);
            strncat(buffer, s, bufsize);
    } else {
        strncat(buffer, "\r\n  Java virtual machine version: (null)", bufsize);
        strncat(buffer, "\r\n  Access Bridge Java class version: (null)", bufsize);
        strncat(buffer, "\r\n  Access Bridge Java DLL version: (null)", bufsize);
    }
    wsprintf(s, "\r\n  Access Bridge Windows DLL version: %ls", versionInfo.bridgeWinDLLVersion);
        strncat(buffer, s, bufsize);

    if (ac == (AccessibleContext) 0) {
        strncat(buffer, "\r\n\r\n(null)", bufsize);
    } else {
		AccessibleRelationSetInfo relationSetInfo;


		//
		// Interface Accessible - our stuff
		//
		strncat(buffer, "\r\n\r\nAccessibleContext information:", bufsize);
		sprintf(s, " [at (%d, %d)]", x, y);
		strncat(buffer, s, bufsize);


		// AccessibleRelationSet
		if (GetAccessibleContextInfo(vmID, ac, &info) == FALSE) {
			strncat(buffer, "\r\n    GetAccessibleContextInfo() failed.", bufsize);
		} else {

				// Core info
            wsprintf(s, "\r\n    Name:  %ls", info.name);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n    Description:  %ls", info.description);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n    Role:  %ls", info.role);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n    Role in en_US locale:  %ls", info.role_en_US);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n    States:  %ls", info.states);
				strncat(buffer, s, bufsize);
			wsprintf(s, "\r\n    States in en_US locale:  %ls", info.states_en_US);
				strncat(buffer, s, bufsize);
			sprintf(s, "\r\n    Index in parent:  %d", info.indexInParent);
				strncat(buffer, s, bufsize);
			sprintf(s, "\r\n    Children Count:  %d", info.childrenCount);
				strncat(buffer, s, bufsize);
			sprintf(s, "\r\n    Bounding Rectangle:  [%d, %d, %d, %d]", 
				info.x, info.y, info.x + info.width, info.y + info.height);
				strncat(buffer, s, bufsize);



            // Get some info about the parent
            parentAC = GetAccessibleParentFromContext(vmID, ac);
            if (parentAC != 0) {
		        if (GetAccessibleContextInfo(vmID, parentAC, &parentInfo) == FALSE) {
			        strncat(buffer, "\r\n    GetAccessibleContextInfo() failed for parent.", bufsize);
		        } else {
			        sprintf(s, "\r\n    Parent's Name:  %ls", parentInfo.name);
				        strncat(buffer, s, bufsize);
                }
                ReleaseJavaObject(vmID, parentAC);
            } else {
			        strncat(buffer, "\r\n    GetAccessibleParentFromContext failed!!!", bufsize);
            }


			if (getAccessibleRelationSet(vmID, ac, &relationSetInfo) == FALSE) {
				strncat(buffer, "\r\nGetAccessibleRelationSet() failed.", bufsize);
			} else {
				int i;
				if (relationSetInfo.relationCount > 0) {
					strncat(buffer, "\r\nAccessibleRelationSet information:", bufsize);
					sprintf(s, " [at (%d, %d)]", x, y);
					strncat(buffer, s, bufsize);
				}
				for (i = 0; i < relationSetInfo.relationCount; i++) {
					AccessibleRelationInfo relationInfo = relationSetInfo.relations[i];

					wsprintf(s, "\r\n  key:  %ls", relationInfo.key);
					strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n  target count:  %d", relationInfo.targetCount);
					strncat(buffer, s, bufsize);
					for (j = 0; j < relationInfo.targetCount; j++) {
						wsprintf(s, "\r\n  target #%d:", j);
						strncat(buffer, s, bufsize);

						if (GetAccessibleContextInfo(vmID, (jobject)relationInfo.targets[j], &relInfo) == FALSE) {
							strncat(buffer, "\r\n    GetAccessibleContextInfo() failed.", bufsize);
						} else {

							// Core info
							wsprintf(s, "\r\n    Target Name:  %ls", relInfo.name);
							strncat(buffer, s, bufsize);
							wsprintf(s, "\r\n    Target Description:  %ls", relInfo.description);
							strncat(buffer, s, bufsize);
							wsprintf(s, "\r\n    Target Role:  %ls", relInfo.role);
							strncat(buffer, s, bufsize);
						}
					}
				}
			}

				// Value info
			if (info.accessibleInterfaces & cAccessibleValueInterface) {

				strncat(buffer, "\r\nAccessible Value information:", bufsize);

				if (GetCurrentAccessibleValueFromContext(vmID, ac, tempBuf, (sizeof(tempBuf) / sizeof(wchar_t))) == TRUE) {
					wsprintf(s, "\r\n    Current Value:  %ls", tempBuf);
				        strncat(buffer, s, bufsize);
				}
				if (GetMaximumAccessibleValueFromContext(vmID, ac, tempBuf, (sizeof(tempBuf) / sizeof(wchar_t))) == TRUE) {
					wsprintf(s, "\r\n    Maximum Value:  %ls", tempBuf);
				        strncat(buffer, s, bufsize);
				}
				if (GetMinimumAccessibleValueFromContext(vmID, ac, tempBuf, (sizeof(tempBuf) / sizeof(wchar_t))) == TRUE) {
					wsprintf(s, "\r\n    Minimum Value:  %ls", tempBuf);
				        strncat(buffer, s, bufsize);
				}
            }

				// Selection info
			if (info.accessibleSelection == TRUE) {

				strncat(buffer, "\r\nAccessible Selection information:", bufsize);

				if ((selCount = GetAccessibleSelectionCountFromContext(vmID, ac)) != -1) {
					wsprintf(s, "\r\n    Selection count:  %d", selCount);
				        strncat(buffer, s, bufsize);

                    for (i = 0; i < selCount; i++) {
				        if ((selectedAC = GetAccessibleSelectionFromContext(vmID, ac, i)) != (jobject) 0) {
		                    if (GetAccessibleContextInfo(vmID, selectedAC, &info) == FALSE) {
                                wsprintf(s, "\r\n    GetAccessibleContextInfo() failed on selection #%d", i);
				                    strncat(buffer, s, bufsize);
		                    } else {
			                    wsprintf(s, "\r\n    Name of selection #%d: %ls", i, info.name);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n    Description of selection #%d: %ls", i, info.description);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n    Role of selection #%d: %ls", i, info.role);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n    States of selection #%d: %ls", i, info.states);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n    Index in parent of selection #%d: %d", i, info.indexInParent);
				                    strncat(buffer, s, bufsize);
			                    wsprintf(s, "\r\n    Children Count of selection #%d: %d", i, info.childrenCount);
				                    strncat(buffer, s, bufsize);
                            }
                            ReleaseJavaObject(vmID, selectedAC);
                        } else {
			                wsprintf(s, "\r\n    GetAccessibleSelectionFromContext() failed on selection #%d", i);
				                    strncat(buffer, s, bufsize);
                        }
				    }
                }
            }

                // Text info
			if (info.accessibleText == TRUE) {
				
				strncat(buffer, "\r\nAccessible Text information:", bufsize);

				if (GetAccessibleTextInfo(vmID, ac, &textInfo, x - info.x, y - info.y) == TRUE) {

					wsprintf(s, "\r\n    Char count:  %d", textInfo.charCount);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n    Caret at index:  %d", textInfo.caretIndex);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n    Index at point:  %d", textInfo.indexAtPoint);
				        strncat(buffer, s, bufsize);

				}
				
                strncat(buffer, "\r\n    Info at point:", bufsize);

				if (GetAccessibleTextRect(vmID, ac, &rectInfo, textInfo.indexAtPoint) == TRUE) {

					wsprintf(s, "\r\n    Character location: [%d, %d], dimension: [%d, %d]",
						rectInfo.x, rectInfo.y, rectInfo.width, rectInfo.height);
				        strncat(buffer, s, bufsize);
				}

                if (GetAccessibleTextItems(vmID, ac, &textItems, textInfo.indexAtPoint) == TRUE) {

					wsprintf(s, "\r\n      Letter:  %lc", textItems.letter);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n      Word:  %ls", textItems.word);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n      Sentence:  %ls", textItems.sentence);
				        strncat(buffer, s, bufsize);

				}

                if (GetAccessibleTextLineBounds(vmID, ac, textInfo.indexAtPoint, &start, &end) == TRUE) {
                    if (GetAccessibleTextRange(vmID, ac, start, end, tempBuf, (sizeof(tempBuf) / sizeof(wchar_t))) == TRUE) {

					    wsprintf(s, "\r\n      **** Line [%d, %d]:  %ls", start, end, tempBuf);
				            strncat(buffer, s, bufsize);

                    }

				}

				if (GetAccessibleTextAttributes(vmID, ac, 
												textInfo.indexAtPoint,
												&attributeInfo)) {

					sprintf(s, "\r\n      Core attributes: %s", 
                        attributeInfo.bold == TRUE ? "bold" : "not bold");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.italic == TRUE ? "italic" : "not italic");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.underline == TRUE ? "underline" : "not underline");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.strikethrough == TRUE ? "strikethrough" : "not strikethrough");
						strncat(buffer, s, bufsize);
					sprintf(s, ",  %s", 
                        attributeInfo.superscript == TRUE ? "superscript" : "not superscript");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.subscript == TRUE ? "subscript" : "not subscript");
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n      Background color:  %ls", attributeInfo.backgroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n      Foreground color:  %ls", attributeInfo.foregroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n      Font family:  %ls", attributeInfo.fontFamily);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Font size:  %d", attributeInfo.fontSize);
						strncat(buffer, s, bufsize);


                    sprintf(s, "\r\n      First line indent:  %f", attributeInfo.firstLineIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Left indent:  %f", attributeInfo.leftIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Right indent:  %f", attributeInfo.rightIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Line spacing:  %f", attributeInfo.lineSpacing);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Space above:  %f", attributeInfo.spaceAbove);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Space below:  %f", attributeInfo.spaceBelow);
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n      Full attribute string:  %ls", attributeInfo.fullAttributesString);
						strncat(buffer, s, bufsize);
				}


				strncat(buffer, "\r\n    Info at caret:", bufsize);   // -------------------------------------

				if (GetAccessibleTextRect(vmID, ac, &rectInfo, textInfo.caretIndex) == TRUE) {

					wsprintf(s, "\r\n    Character location: [%d, %d], dimension: [%d, %d]",
						rectInfo.x, rectInfo.y, rectInfo.width, rectInfo.height);
				        strncat(buffer, s, bufsize);
				}

                if (GetAccessibleTextItems(vmID, ac, &textItems, textInfo.caretIndex) == TRUE) {

					wsprintf(s, "\r\n      Letter:  %lc", textItems.letter);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n      Word:  %ls", textItems.word);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n      Sentence:  %ls", textItems.sentence);
				        strncat(buffer, s, bufsize);

				}

                if (GetAccessibleTextLineBounds(vmID, ac, textInfo.caretIndex, &start, &end) == TRUE) {
                    if (GetAccessibleTextRange(vmID, ac, start, end, tempBuf, (sizeof(tempBuf) / sizeof(wchar_t))) == TRUE) {

					    wsprintf(s, "\r\n      Line [%d, %d]:  %ls", start, end, tempBuf);
				            strncat(buffer, s, bufsize);

                    }

				}

                if (GetAccessibleTextAttributes(vmID, ac, 
												textInfo.caretIndex,
                                                &attributeInfo)) {

					sprintf(s, "\r\n      Core attributes: %s", 
                        attributeInfo.bold == TRUE ? "bold" : "not bold");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.italic == TRUE ? "italic" : "not italic");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.underline == TRUE ? "underline" : "not underline");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.strikethrough == TRUE ? "strikethrough" : "not strikethrough");
						strncat(buffer, s, bufsize);
					sprintf(s, ",  %s", 
                        attributeInfo.superscript == TRUE ? "superscript" : "not superscript");
						strncat(buffer, s, bufsize);
					sprintf(s, ", %s", 
                        attributeInfo.subscript == TRUE ? "subscript" : "not subscript");
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n      Background color:  %ls", attributeInfo.backgroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n      Foreground color:  %ls", attributeInfo.foregroundColor);
						strncat(buffer, s, bufsize);
                    wsprintf(s, "\r\n      Font family:  %ls", attributeInfo.fontFamily);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Font size:  %d", attributeInfo.fontSize);
						strncat(buffer, s, bufsize);


                    sprintf(s, "\r\n      First line indent:  %f", attributeInfo.firstLineIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Left indent:  %f", attributeInfo.leftIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Right indent:  %f", attributeInfo.rightIndent);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Line spacing:  %f", attributeInfo.lineSpacing);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Space above:  %f", attributeInfo.spaceAbove);
						strncat(buffer, s, bufsize);
                    sprintf(s, "\r\n      Space below:  %f", attributeInfo.spaceBelow);
						strncat(buffer, s, bufsize);

                    wsprintf(s, "\r\n      Full attribute string:  %ls", attributeInfo.fullAttributesString);
						strncat(buffer, s, bufsize);
				}

                if (GetAccessibleTextSelectionInfo(vmID, ac, &textSelection) == TRUE) {

					wsprintf(s, "\r\n    Selection start index:  %d", textSelection.selectionStartIndex);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n    Selection end index:  %d", textSelection.selectionEndIndex);
				        strncat(buffer, s, bufsize);
					wsprintf(s, "\r\n    Selected text:  %ls", textSelection.selectedText);
				        strncat(buffer, s, bufsize);

				}
            }
		}
	}

    return buffer;													  
}


/**
 * Display Accessible information about the object under the mouse
 */
void displayAccessibleInfo(long vmID, AccessibleContext ac, int x, int y) {
	char buffer[TEXTBUFSIZE];

//    return;

	getAccessibleInfo(vmID, ac, x, y, buffer, TEXTBUFSIZE);
	SetDlgItemText(theDialogWindow, cFerretText, buffer);														  
}


/**
 * Display Accessible propertyChange event info
 */
void displayAccessiblePropertyChange(long vmID, AccessibleContext ac,
							         char *announcement) {
	char buffer[TEXTBUFSIZE];
	char *bufOffset;

	strncpy(buffer, announcement, TEXTBUFSIZE);

	bufOffset = (char *) (buffer + strlen(buffer));
	getAccessibleInfo(vmID, ac, 0, 0, bufOffset, TEXTBUFSIZE - strlen(buffer));

	SetDlgItemText(theDialogWindow, cFerretText, buffer);														  
}


/**
 * Update display under mouse when it moves
 */
void echoMouseObject() {
	long vmID;
	AccessibleContext acParent;
	AccessibleContext ac;
	POINT p;
	HWND hwnd;
	RECT windowRect;

	GetCursorPos(&p);
	hwnd = WindowFromPoint(p);
	if (GetAccessibleContextFromHWND(hwnd, &vmID, &acParent)) {
		GetWindowRect(hwnd, &windowRect);
			// send the point in global coordinates; Java will handle it!
		if (GetAccessibleContextAt(vmID, acParent, (jint) p.x, (jint) p.y, &ac)) {
			displayAccessibleInfo(vmID, ac, p.x, p.y);		// can handle null
			ReleaseJavaObject(vmID, ac);
		}
	}
}


/**
 * Update display under HWND the mouse is in
 */
void echoMouseHWNDObject() {
	long vmID;
	AccessibleContext ac;
	POINT p;
	HWND hwnd;

	GetCursorPos(&p);
	hwnd = WindowFromPoint(p); 

	if (GetAccessibleContextFromHWND(hwnd, &vmID, &ac)) {
		displayAccessibleInfo(vmID, ac, 0, 0);		// can handle null
		ReleaseJavaObject(vmID, ac);
	}
}

/**
 * Display Accessible information about the object that has focus in 
 * the topmost Java HWND
 *
 */
void displayFocusedObject() {
	HWND hWnd;
	long vmID;
	AccessibleContext ac;

	hWnd = GetTopWindow(NULL);
	while (hWnd != NULL) {
		if (IsJavaWindow(hWnd)) {
			if (GetAccessibleContextWithFocus(hWnd, &vmID, &ac) == TRUE) {
				displayAccessibleInfo(vmID, ac, 0, 0);
            	ReleaseJavaObject(vmID, ac);
			}
			return;
		} else {
			hWnd = GetNextWindow(hWnd, GW_HWNDNEXT);
		}
	}
}


/**
 * Handle a FocusGained event
 */
void HandleJavaFocusGained(long vmID, FocusEvent event, AccessibleContext ac) {

    // ignore passed in AccessibleContext, in order to test the
    // GetAccessibleContextWithFocus() function
//    displayFocusedObject();  -- bad practice, as you may be asking the wrong JavaVM!

    if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a FocusLost event
 */
void HandleJavaFocusLost(long vmID, FocusEvent event, AccessibleContext ac) {

    // NOTE: calling GetAccessibleContextWithFocus() after a FocusLost event
    //       would result in a null AccessibleContext being returned, since
    //       at that point, no object has the focus.  If the topmost window
    //       does not belong to a JavaVM, then no component within a JavaVM
    //       will have the focus (and again, GetAccessibleContextWithFocus()
    //       will return a null AccessibleContext).  You should always get
    //       a FocusLost event when a window not belonging to a JavaVM becomes
    //       topmost.

	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a CaretUpdate event
 */
void HandleJavaCaretUpdate(long vmID, CaretEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MouseClicked event
 */
void HandleMouseClicked(long vmID, MouseEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MouseEntered event
 */
void HandleMouseEntered(long vmID, MouseEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}

	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MouseExited event
 */
void HandleMouseExited(long vmID, MouseEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MousePressed event
 */
void HandleMousePressed(long vmID, MouseEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MouseReleased event
 */
void HandleMouseReleased(long vmID, MouseEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MenuCanceled event
 */
void HandleMenuCanceled(long vmID, MenuEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MenuDeselected event
 */
void HandleMenuDeselected(long vmID, MenuEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a MenuSelected event
 */
void HandleMenuSelected(long vmID, MenuEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a PopupMenuCanceled event
 */
void HandlePopupMenuCanceled(long vmID, MenuEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a PopupMenuWillBecomeInvisible event
 */
void HandlePopupMenuWillBecomeInvisible(long vmID, MenuEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a PopupMenuWillBecomeVisible event
 */
void HandlePopupMenuWillBecomeVisible(long vmID, MenuEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessibleInfo(vmID, ac, 0, 0);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}




/**
 * Handle a HandlePropertyNameChange event
 */
void HandlePropertyNameChange(long vmID, PropertyChangeEvent event, AccessibleContext ac,
                               wchar_t *oldName, wchar_t *newName) {
	char s[1024];
    wsprintf(s, "Name changed event: old = %ls; new = %ls\r\n\r\n", oldName, newName);

	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, s);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyDescriptionChange event
 */
void HandlePropertyDescriptionChange(long vmID, PropertyChangeEvent event, AccessibleContext ac,
                                     wchar_t *oldDescription, wchar_t *newDescription) {
	char s[1024];
    wsprintf(s, "Description changed event: old = %ls; new = %ls\r\n\r\n", oldDescription, newDescription);

	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, s);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyStateChange event
 */
void HandlePropertyStateChange(long vmID, PropertyChangeEvent event, AccessibleContext ac,
                               wchar_t *oldState, wchar_t *newState) {
	char s[1024];
    wsprintf(s, "State changed event: old = %ls; new = %ls\r\n\r\n", oldState, newState);

	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, s);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyValueChange event
 */
void HandlePropertyValueChange(long vmID, PropertyChangeEvent event, AccessibleContext ac,
                               wchar_t *oldValue, wchar_t *newValue) {
	char s[1024];
    wsprintf(s, "Value changed event: old = %ls; new = %ls\r\n\r\n", oldValue, newValue);

	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, s);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertySelectionChange event
 */
void HandlePropertySelectionChange(long vmID, PropertyChangeEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, "Selection changed event\r\n\r\n");
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyTextChange event
 */
void HandlePropertyTextChange(long vmID, PropertyChangeEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, "Text changed event\r\n\r\n");
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyCaretChange event
 */
void HandlePropertyCaretChange(long vmID, PropertyChangeEvent event, AccessibleContext ac,
                               int oldPosition, int newPosition) {
	char s[1024];

    wsprintf(s, "Caret changed event: oldPosition = %d; newPosition = %d\r\n\r\n", oldPosition, newPosition);

	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, s);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyVisibleDataChange event
 */
void HandlePropertyVisibleDataChange(long vmID, PropertyChangeEvent event, AccessibleContext ac) {
	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, "VisibleData changed event\r\n\r\n");
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyChildChange event
 */
void HandlePropertyChildChange(long vmID, PropertyChangeEvent event, AccessibleContext ac,
                               jobject oldChild, jobject newChild) {
	char buffer[TEXTBUFSIZE];
	char *bufOffset;

    sprintf(buffer, "Child property changed event:\r\n=======================\r\n\r\n");

    if (oldChild != 0) {
        strncat(buffer, "Old Accessible Child info:\r\n\r\n", TEXTBUFSIZE);
        bufOffset = (char *) (buffer + strlen(buffer));
	    getAccessibleInfo(vmID, oldChild, 0, 0, bufOffset, TEXTBUFSIZE - strlen(buffer));
        strncat(buffer, "\r\n\r\n", TEXTBUFSIZE);
    }

    if (newChild != 0) {
        strncat(buffer, "New Accessible Child info:\r\n\r\n", TEXTBUFSIZE);
        bufOffset = (char *) (buffer + strlen(buffer));
	    getAccessibleInfo(vmID, newChild, 0, 0, bufOffset, TEXTBUFSIZE - strlen(buffer));
        strncat(buffer, "\r\n\r\n", TEXTBUFSIZE);
    }
    
    if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, buffer);
	}

	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, oldChild);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, newChild);	// must always release, to stave off memory leaks
}

/**
 * Handle a HandlePropertyActiveDescendentChange event
 */
void HandlePropertyActiveDescendentChange(long vmID, PropertyChangeEvent event, 
                                          AccessibleContext ac, 
                                          jobject oldActiveDescendent,
                                          jobject newActiveDescendent) {
	char buffer[TEXTBUFSIZE];
	char *bufOffset;

    sprintf(buffer, "ActiveDescendent property changed event:\r\n=======================\r\n\r\n");

    if (oldActiveDescendent != 0) {
        strncat(buffer, "Old Accessible ActiveDescendent info:\r\n\r\n", TEXTBUFSIZE);
        bufOffset = (char *) (buffer + strlen(buffer));
	    getAccessibleInfo(vmID, oldActiveDescendent, 0, 0, bufOffset, TEXTBUFSIZE - strlen(buffer));
        strncat(buffer, "\r\n\r\n", TEXTBUFSIZE);
    }

    if (newActiveDescendent != 0) {
        strncat(buffer, "New Accessible ActiveDescendent info:\r\n\r\n", TEXTBUFSIZE);
        bufOffset = (char *) (buffer + strlen(buffer));
	    getAccessibleInfo(vmID, newActiveDescendent, 0, 0, bufOffset, TEXTBUFSIZE - strlen(buffer));
        strncat(buffer, "\r\n\r\n", TEXTBUFSIZE);
    }
        
	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, buffer);
	}

	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, oldActiveDescendent);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, newActiveDescendent);	// must always release, to stave off memory leaks
}


/**
 * Handle a HandlePropertyTableModelChange event
 */
void HandlePropertyTableModelChange(long vmID, PropertyChangeEvent event, 
                                    AccessibleContext ac, 
                                    wchar_t *oldValue, wchar_t *newValue) {

	char s[1024];
    wsprintf(s, "Table Model Change: old = %ls; new = %ls\r\n\r\n", oldValue, newValue);

	if (ac != (AccessibleContext) 0) {
		displayAccessiblePropertyChange(vmID, ac, s);
	}
	ReleaseJavaObject(vmID, ac);	// must always release, to stave off memory leaks
	ReleaseJavaObject(vmID, event);	// must always release, to stave off memory leaks
}



#define DOWN_UP_FLAG 1<<31

#ifdef __cplusplus
extern "C" {
#endif
LRESULT CALLBACK KeyboardProc(int code, WPARAM wParam, LPARAM lParam) {
		// on mouse-up of F1
	if (code < 0) {
		CallNextHookEx(prevKbdHook, code, wParam, lParam);
	} else if (wParam == VK_F1 && lParam & DOWN_UP_FLAG) {
		PostMessage(theDialogWindow, DISPLAY_INFO_MESSAGE, wParam, lParam);
	} else if (wParam == VK_F2 && lParam & DOWN_UP_FLAG) {
		PostMessage(theDialogWindow, DISPLAY_HWND_INFO_MESSAGE, wParam, lParam);
	}
	return 0;
}

LRESULT CALLBACK MouseProc(int code, WPARAM wParam, LPARAM lParam) {
		// when mouse settles from movement
	if (code < 0) {
		CallNextHookEx(prevMouseHook, code, wParam, lParam);
	} else {
			// reset the timer every time the mouse moves
		KillTimer(theDialogWindow, TIMER_ID);
		SetTimer(theDialogWindow, TIMER_ID, 1000, TimerProc);
	}
	return 0;
}

LRESULT CALLBACK TimerProc(HWND hWnd, UINT uTimerMsg, UINT uTimerID, DWORD dwTime) {
		// when mouse settles from movement
	KillTimer(hWnd, uTimerID);
	if (updateMouse == TRUE) {
		PostMessage(theDialogWindow, DISPLAY_INFO_MESSAGE, 0, 0);
	}
	return 0;
}

#ifdef __cplusplus
}
#endif

void exitJavaFerret(HWND hWnd) {
	EndDialog(hWnd, TRUE);
	PostQuitMessage (0);
}

#define INSTALL_EVENT_LISTENER(toggleVal, setFP, handler)   \
    if (toggleVal) {                                        \
        setFP(handler);                                     \
    }

void reinstallEventListeners() {
	INSTALL_EVENT_LISTENER(trackMouse, SetMouseEntered, HandleMouseEntered);
	INSTALL_EVENT_LISTENER(trackFocus, SetFocusGained, HandleJavaFocusGained);
	INSTALL_EVENT_LISTENER(trackCaret, SetCaretUpdate, HandleJavaCaretUpdate);

	INSTALL_EVENT_LISTENER(trackMenuSelected, SetMenuSelected, HandleMenuSelected);
	INSTALL_EVENT_LISTENER(trackMenuDeselected, SetMenuDeselected, HandleMenuDeselected);
	INSTALL_EVENT_LISTENER(trackMenuCanceled, SetMenuCanceled, HandleMenuCanceled);

	INSTALL_EVENT_LISTENER(trackPopupVisible, SetPopupMenuWillBecomeVisible, HandlePopupMenuWillBecomeVisible);
	INSTALL_EVENT_LISTENER(trackPopupInvisible, SetPopupMenuWillBecomeInvisible, HandlePopupMenuWillBecomeInvisible);
	INSTALL_EVENT_LISTENER(trackPopupCanceled, SetPopupMenuCanceled, HandlePopupMenuCanceled);

	INSTALL_EVENT_LISTENER(trackPropertyNameChange, SetPropertyNameChange, HandlePropertyNameChange);
	INSTALL_EVENT_LISTENER(trackPropertyDescriptionChange, SetPropertyDescriptionChange, HandlePropertyDescriptionChange);
	INSTALL_EVENT_LISTENER(trackPropertyStateChange, SetPropertyStateChange, HandlePropertyStateChange);
	INSTALL_EVENT_LISTENER(trackPropertyValueChange, SetPropertyValueChange, HandlePropertyValueChange);
	INSTALL_EVENT_LISTENER(trackPropertySelectionChange, SetPropertySelectionChange, HandlePropertySelectionChange);
	INSTALL_EVENT_LISTENER(trackPropertyTextChange, SetPropertyTextChange, HandlePropertyTextChange);
	INSTALL_EVENT_LISTENER(trackPropertyCaretChange, SetPropertyCaretChange, HandlePropertyCaretChange);
	INSTALL_EVENT_LISTENER(trackPropertyVisibleDataChange, SetPropertyVisibleDataChange, HandlePropertyVisibleDataChange);
	INSTALL_EVENT_LISTENER(trackPropertyChildChange, SetPropertyChildChange, HandlePropertyChildChange);
	INSTALL_EVENT_LISTENER(trackPropertyActiveDescendentChange, SetPropertyActiveDescendentChange, HandlePropertyActiveDescendentChange);
	INSTALL_EVENT_LISTENER(trackPropertyTableModelChange, SetPropertyTableModelChange, HandlePropertyTableModelChange);
}


#define TRACK_EVENT_TOGGLE(menuItem, toggleVal, setFP, handler)	\
	case menuItem:												\
		menu = GetMenu(hWnd);									\
		if (toggleVal) {										\
			toggleVal = FALSE;									\
			CheckMenuItem(menu, menuItem,						\
						  MF_BYCOMMAND | MF_UNCHECKED);			\
			setFP(NULL);										\
		} else {												\
			toggleVal = TRUE;									\
			CheckMenuItem(menu, menuItem,						\
						  MF_BYCOMMAND | MF_CHECKED);			\
			setFP(handler);										\
		}														\
		return(TRUE)


#ifdef __cplusplus
extern "C" {
#endif
BOOL APIENTRY FerretDialogProc (HWND hWnd, UINT message, UINT wParam, LONG lParam) {
	int command;
	short width, height;
	HWND dlgItem;
	HWND menu;

	switch (message) {
	case WM_INITDIALOG:
/*
        if (theAccessBridgeLoadedFlag) {
			CheckMenuItem(GetMenu(hWnd), cAccessBridgeDLLLoaded, MF_BYCOMMAND | MF_CHECKED);
        } else {
			CheckMenuItem(GetMenu(hWnd), cAccessBridgeDLLLoaded, MF_BYCOMMAND | MF_UNCHECKED);
        }
*/
            // initWindow() happens before theAccessBridgeLoadedFlag is set; assume it gets set to true
        CheckMenuItem(GetMenu(hWnd), cAccessBridgeDLLLoaded, MF_BYCOMMAND | MF_CHECKED);
		return(TRUE);

	case WM_CLOSE:
		exitJavaFerret(hWnd);
		return(TRUE);
		break;
	
	case WM_SIZE:
		width = LOWORD(lParam);
		height = HIWORD(lParam);
		dlgItem = GetDlgItem(theDialogWindow, cFerretText); 
		SetWindowPos(dlgItem, NULL, 0, 0, width, height, 0); 
		return(FALSE);			// let windows finish handling this
		break;
	
	case WM_COMMAND:
		command = LOWORD(wParam);
		switch(command) {
        case cAccessBridgeDLLLoaded:    // toggle; unload or load AccessBridge
            if (theAccessBridgeLoadedFlag) {
                shutdownAccessBridge();
                theAccessBridgeLoadedFlag = FALSE;
			    CheckMenuItem(GetMenu(hWnd), cAccessBridgeDLLLoaded, MF_BYCOMMAND | MF_UNCHECKED);
            } else {
                theAccessBridgeLoadedFlag = initializeAccessBridge();
                if (theAccessBridgeLoadedFlag) {
			        CheckMenuItem(GetMenu(hWnd), cAccessBridgeDLLLoaded, MF_BYCOMMAND | MF_CHECKED);
                    reinstallEventListeners();
                }
            }
            return(TRUE);

        case cExitMenuItem:
			exitJavaFerret(hWnd);
			return(TRUE);


		TRACK_EVENT_TOGGLE(cTrackMouseMenuItem, trackMouse, SetMouseEntered, HandleMouseEntered);
		TRACK_EVENT_TOGGLE(cTrackFocusMenuItem, trackFocus, SetFocusGained, HandleJavaFocusGained);
		TRACK_EVENT_TOGGLE(cTrackCaretMenuItem, trackCaret, SetCaretUpdate, HandleJavaCaretUpdate);

		TRACK_EVENT_TOGGLE(cTrackMenuSelectedMenuItem, trackMenuSelected, SetMenuSelected, HandleMenuSelected);
		TRACK_EVENT_TOGGLE(cTrackMenuDeselectedMenuItem, trackMenuDeselected, SetMenuDeselected, HandleMenuDeselected);
		TRACK_EVENT_TOGGLE(cTrackMenuCanceledItem, trackMenuCanceled, SetMenuCanceled, HandleMenuCanceled);

		TRACK_EVENT_TOGGLE(cTrackPopupBecomeVisibleMenuItem, trackPopupVisible, SetPopupMenuWillBecomeVisible, HandlePopupMenuWillBecomeVisible);
		TRACK_EVENT_TOGGLE(cTrackPopupBecomeInvisibleMenuItem, trackPopupInvisible, SetPopupMenuWillBecomeInvisible, HandlePopupMenuWillBecomeInvisible);
		TRACK_EVENT_TOGGLE(cTrackPopupCanceledItem, trackPopupCanceled, SetPopupMenuCanceled, HandlePopupMenuCanceled);

		TRACK_EVENT_TOGGLE(cTrackPropertyNameItem, trackPropertyNameChange, SetPropertyNameChange, HandlePropertyNameChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyDescriptionItem, trackPropertyDescriptionChange, SetPropertyDescriptionChange, HandlePropertyDescriptionChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyStateItem, trackPropertyStateChange, SetPropertyStateChange, HandlePropertyStateChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyValueItem, trackPropertyValueChange, SetPropertyValueChange, HandlePropertyValueChange);
		TRACK_EVENT_TOGGLE(cTrackPropertySelectionItem, trackPropertySelectionChange, SetPropertySelectionChange, HandlePropertySelectionChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyTextItem, trackPropertyTextChange, SetPropertyTextChange, HandlePropertyTextChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyCaretItem, trackPropertyCaretChange, SetPropertyCaretChange, HandlePropertyCaretChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyVisibleDataItem, trackPropertyVisibleDataChange, SetPropertyVisibleDataChange, HandlePropertyVisibleDataChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyChildItem, trackPropertyChildChange, SetPropertyChildChange, HandlePropertyChildChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyActiveDescendentItem, trackPropertyActiveDescendentChange, SetPropertyActiveDescendentChange, HandlePropertyActiveDescendentChange);
		TRACK_EVENT_TOGGLE(cTrackPropertyTableModelChangeItem, trackPropertyTableModelChange, SetPropertyTableModelChange, HandlePropertyTableModelChange);
	

		case cUpdateFromMouseMenuItem:
			menu = GetMenu(hWnd);
			if (updateMouse) {
				updateMouse = FALSE;
				CheckMenuItem(menu, cUpdateFromMouseMenuItem, 
							  MF_BYCOMMAND | MF_UNCHECKED);
				UnhookWindowsHookEx(MouseProc);
				KillTimer(hWnd, TIMER_ID);
			} else {
				updateMouse = TRUE;
				CheckMenuItem(menu, cUpdateFromMouseMenuItem, 
							  MF_BYCOMMAND | MF_CHECKED);
				prevMouseHook = SetWindowsHookEx(WH_MOUSE, MouseProc, theInstance, 0);
			}
			return(TRUE);

		case cUpdateWithF1Item:
			menu = GetMenu(hWnd);
			if (updateF1) {
				updateF1 = FALSE;
				CheckMenuItem(menu, cUpdateWithF1Item,  
							  MF_BYCOMMAND | MFS_UNCHECKED);
				UnhookWindowsHookEx(KeyboardProc);
			} else {
				updateF1 = TRUE;
				CheckMenuItem(menu, cUpdateWithF1Item,  
							  MF_BYCOMMAND | MFS_CHECKED);
				prevKbdHook = SetWindowsHookEx(WH_KEYBOARD, KeyboardProc, theInstance, 0);
			}
			return(TRUE);

		case cUpdateWithF2Item:
			menu = GetMenu(hWnd);
			if (updateF2) {
				updateF2 = FALSE;
				CheckMenuItem(menu, cUpdateWithF2Item,  
							  MF_BYCOMMAND | MFS_UNCHECKED);
				UnhookWindowsHookEx(KeyboardProc);
			} else {
				updateF2 = TRUE;
				CheckMenuItem(menu, cUpdateWithF2Item,  
							  MF_BYCOMMAND | MFS_CHECKED);
				prevKbdHook = SetWindowsHookEx(WH_KEYBOARD, KeyboardProc, theInstance, 0);
			}
			return(TRUE);

		}
		break;

	case DISPLAY_INFO_MESSAGE:
		echoMouseObject();
		break;

	case DISPLAY_HWND_INFO_MESSAGE:
		echoMouseHWNDObject();
		break;
	}

	return (FALSE);   
}
#ifdef __cplusplus
}
#endif
